#!/usr/libexec/platform-python

"""
	Moon Lander - Horrible homage to Lunar Lander

    Copyright (C) 2021 Joshua Boudreau <jboudreau@45drives.com>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from FFDHud import FFDHud
import sys
from PIL import Image
import time
import math
from typing import Union
import numpy as np

class Entity:
	def __init__(self, hud : FFDHud, sprite_path : str, pos : (float, float), vel : (float, float),
			acc : (float, float) = (0.0,0.0)):
		self.pos = np.array(pos).astype(float)
		self.vel = np.array(vel).astype(float)
		self.acc = np.array(acc).astype(float)
		self.sprite = Image.open(sprite_path)
		self.sprite_corner_offset = np.array(self.sprite.size) / 2
		self.x_lim = float(hud.sizex)
		self.y_lim = float(hud.sizey)
		self.diagonal_size = math.ceil(math.sqrt(float(self.sprite.size[0]) ** 2.0 + float(self.sprite.size[1]) ** 2.0))
	
	def __del__(self):
		self.sprite.close()
	
	def interact(self, other, game):
		pass
	
	def update(self, timedelta : float, game):
		self.vel += self.acc * timedelta
		self.pos += self.vel * timedelta
	
	def get_angle(self):
		return 0
	
	def draw(self, hud : FFDHud):
		sprite = Image.new(self.sprite.mode, (self.diagonal_size, self.diagonal_size))
		paste_x = (self.diagonal_size - self.sprite.size[0])//2
		paste_y = (self.diagonal_size - self.sprite.size[1])//2
		sprite.paste(self.sprite, (paste_x, paste_y, paste_x + self.sprite.size[0], paste_y + self.sprite.size[1]))
		sprite_corner_offset = np.array(sprite.size) / 2
		hud.graphics_buffer_blit(sprite.rotate(self.get_angle()),
								 tuple(((self.pos) * (1, -1) + (0.0, hud.sizey) - sprite_corner_offset).astype(int).tolist()),
								 sprite.size,
								 "add")

	def top_edge(self):
		return self.pos[1] + self.sprite.size[1] / 2
	def bottom_edge(self):
		return self.pos[1] - self.sprite.size[1] / 2
	def left_edge(self):
		return self.pos[0] - self.sprite.size[0] / 2
	def right_edge(self):
		return self.pos[0] + self.sprite.size[0] / 2

	def check_hit(self, other):
		return(
			other.right_edge() >= self.left_edge()
			and other.left_edge() <= self.right_edge()
			and other.top_edge() >= self.bottom_edge()
			and other.bottom_edge() <= self.top_edge()
		)

class Player(Entity):
	def __init__(self, hud : FFDHud, sprite_path : str, thrust_sprite_path : str, pos : (float, float), vel : (float, float),
			acc : (float, float) = (0.0,0.0)):
		Entity.__init__(self, hud, sprite_path, pos, vel, acc)
		self.thrust_sprite = Image.open(thrust_sprite_path)
		self.thrust_sprite_corner_offset = np.array(self.thrust_sprite.size) / 2
		self.ext_acc = np.array((0.0,0.0))
		self.direction = np.array((0.0,1.0))
		self.hud = hud
		self.fuel = 100.0

	def draw(self, hud : FFDHud):
		Entity.draw(self, hud)
		if np.linalg.norm(self.ext_acc):
			hud.graphics_buffer_blit(self.thrust_sprite,
				tuple(((self.pos - self.direction * 7) * (1, -1) + (0.0, hud.sizey) - self.thrust_sprite_corner_offset).astype(int).tolist()),
				self.thrust_sprite.size,
				"add")
	
	def thrust(self, scalar : float):
		self.ext_acc = self.direction * scalar
	
	def rotate(self, theta : float):
		rot = np.array([[math.cos(theta), -math.sin(theta)], [math.sin(theta), math.cos(theta)]])
		self.direction = np.dot(rot, self.direction)
	
	def get_angle(self):
		angle = np.arccos(self.direction @ (1.0, 0.0)) * 360.0 / (2.0 * math.pi)
		if self.direction[1] < 0.0:
			angle = 360.0 - angle
		return angle
	
	def interact(self, other : Entity, game):
		if np.linalg.norm(self.vel - other.vel) > 10:
			game.result = "Crash! (too fast)\nGame Over"
			return
		if isinstance(other, Goal):
			angle = self.get_angle()
			if self.left_edge() < other.left_edge() or self.right_edge() > other.right_edge():
				game.result = "Crash! (not fully on pad)\nGame Over"
			elif angle < 45 or angle > 135:
				game.result = "Crash! (a little sideways there bud)\nGame Over"
			else:
				game.result = f"Win!\nScore: {round(self.fuel)}"
	
	def update(self, timedelta : float, game):
		for event in self.hud.get_encoder_events():
			if event[0]:
				self.rotate(event[0] * 0.261799)
			if event[1] and self.fuel > 0.0:
				self.thrust(10.0)
			else:
				self.thrust(0.0)
		if np.linalg.norm(self.ext_acc):
			self.fuel = max(self.fuel - timedelta * 5.0, 0.0)
		self.acc += self.ext_acc
		Entity.update(self, timedelta, game)
		self.acc -= self.ext_acc
		if self.right_edge() < 0.0 or self.left_edge() >= float(self.hud.sizex) or self.bottom_edge() >= float(self.hud.sizey):
			game.result = "Out of bounds!\nGame Over"
		if self.bottom_edge()< 0:
			game.result = "Crash!\nGame Over"

class Goal(Entity):
	def __init__(self, hud : FFDHud, sprite_path : str, pos : (float, float), vel : (float, float),
			acc : (float, float) = (0.0,0.0)):
		Entity.__init__(self, hud, sprite_path, pos, vel, acc)

class Game:
	def __init__(self, hud : FFDHud):
		self.hud = hud
		self.entities = []
		self.result = None
	
	def loop(self):
		timestamp = time.time()
		while not self.result:
			now = time.time()
			timedelta = now - timestamp
			timestamp = now
			self.hud.graphics_buffer_clear()
			for entity1 in self.entities:
				for entity2 in self.entities:
					if entity1 == entity2:
						continue
					if entity1.check_hit(entity2):
						entity1.interact(entity2, self)
				entity1.update(timedelta, self)
				entity1.draw(self.hud)
			self.hud.graphics_buffer_show()
			self.hud.move_cursor_abs(self.hud.rows, 1)
			self.hud.print(f"Fuel: {round(self.entities[0].fuel, 1)} {'SPEED' if np.linalg.norm(self.entities[0].vel) > 10 else ''}")
		self.hud.clear()
		self.hud.print(self.result)
		time.sleep(4)
	
	def add_entity(self, ent : Entity):
		self.entities.append(ent)

def main(dev="45D_HUD1"):
	HUD = FFDHud("/dev/" + dev)
	HUD.clear()
	game = Game(HUD)
	game.add_entity(Player(HUD, "game_assets/ship.png", "game_assets/thrust.png", (5, HUD.sizey - 5), (0.0, 0.0), (0.0, -5.0)))
	game.add_entity(Goal(HUD, "game_assets/paddle.png", (HUD.sizex - 15, 5), (0.0, 0.0), (0.0, 0.0)))
	game.loop()

if __name__ == "__main__":
	main()
