#!/usr/libexec/platform-python

"""
	Serpent - Horrible homage to Snake

    Copyright (C) 2021 Joshua Boudreau <jboudreau@45drives.com>
    
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from FFDHud import FFDHud, FFDHudMenu
import sys
from PIL import Image
import time
from typing import Union
import random
from functools import partial

class Entity:
	def __init__(self, sprite : Image.Image, pos : (int, int)):
		self.pos = [*pos]
		self.sprite = sprite
		self.sprite_corner_offset = [sprite.size[0] // 2, sprite.size[1] // 2]	
	def draw(self, hud : FFDHud):
		hud.graphics_buffer_blit(self.sprite,
								 tuple([self.pos[0] - self.sprite_corner_offset[0], self.pos[1] - self.sprite_corner_offset[1]]),
								 self.sprite.size,
								 "add")

class Segment(Entity):
	def __init__(self, sprite : Image.Image, pos : (int, int)):
		Entity.__init__(self, sprite, pos)

class Snake:
	def __init__(self, pos : (int, int), dir : (int, int)):
		segment_sprite = Image.open("game_assets/serpent_goal.png")
		segment_diameter = segment_sprite.size[0]
		pos = [(pos[0] // segment_diameter) * segment_diameter, (pos[1] // segment_diameter) * segment_diameter]
		self.dir = [dir[0] * segment_sprite.size[0], dir[1] * segment_sprite.size[1]]
		self.segments = [Segment(segment_sprite, pos)]

	def advance(self, game):
		head = self.segments[0]
		test = [head.pos[0] + self.dir[0], head.pos[1] + self.dir[1]]
		if head.pos[0] <= 0 or head.pos[0] >= game.hud.sizex or head.pos[1] <= 0 or head.pos[1] >= game.hud.sizey:
			game.running = False
		growth = None
		for goal in game.goals:
			if goal.pos[0] == test[0] and goal.pos[1] == test[1]:
				game.score += 1
				growth = Segment(self.segments[-1].sprite, self.segments[-1].pos)
				game.goals.remove(goal)
				game.new_goal()
		for i in reversed(range(len(self.segments) - 1)):
			n0 = self.segments[i]
			n1 = self.segments[i+1]
			n1.pos[0] = n0.pos[0]
			n1.pos[1] = n0.pos[1]
			if n0.pos[0] == test[0] and n0.pos[1] == test[1]:
				game.running = False
		head.pos = test
		if growth:
			self.segments.append(growth)

	def turn(self, rot : int):
		sintheta = -rot
		self.dir = [-self.dir[1] * sintheta, self.dir[0] * sintheta]

	def draw(self, hud):
		for segment in self.segments:
			segment.draw(hud)

class Game:
	def __init__(self, hud : FFDHud):
		self.running = True
		self.goals = []
		self.hud = hud
		self.snake = Snake((hud.sizex // 2, hud.sizey // 2), (0, -1))
		self.score = 0
		self.goal_sprite = Image.open("game_assets/serpent_goal.png")
		self.walls = []
		wall_sprite = Image.open("game_assets/serpent_goal.png")
		for i in range(0, hud.sizex, wall_sprite.size[0]):
			self.walls.append(Entity(wall_sprite, (i, 0)))
			self.walls.append(Entity(wall_sprite, (i, hud.sizey)))
		for j in range(0, hud.sizey, wall_sprite.size[1]):
			self.walls.append(Entity(wall_sprite, (0, j)))
			self.walls.append(Entity(wall_sprite, (hud.sizex, j)))

	def new_goal(self):
		snake_sections = [segment.pos for segment in self.snake.segments]
		goal_diameter = self.goal_sprite.size[0]
		goal_radius = goal_diameter // 2
		allowed_sections = []
		for i in range(goal_diameter, self.hud.sizex, goal_diameter):
			for j in range(goal_diameter, self.hud.sizey, goal_diameter):
				collision = False
				for snake_section in snake_sections:
					if i == snake_section[0] and j == snake_section[1]:
						collision = True
						break
				if not collision:
					allowed_sections.append([i, j])
		ind = random.randint(0, len(allowed_sections) - 1)
		self.goals.append(Entity(self.goal_sprite, tuple(allowed_sections[ind])))

	def loop(self, frame_period):
		self.new_goal()
		while self.running:
			self.hud.graphics_buffer_clear()
			self.snake.advance(self)
			self.snake.draw(self.hud)
			for goal in self.goals:
				goal.draw(self.hud)
			for wall in self.walls:
				wall.draw(self.hud)
			self.hud.graphics_buffer_show()
			time.sleep(frame_period)
			events = self.hud.get_encoder_events()
			for event in events:
				if event[0]:
					self.snake.turn(event[0])
					break
		self.hud.clear()
		self.hud.move_cursor_abs(1, 1)
		self.hud.print(f"Score: {self.score}")
		time.sleep(4)

def get_frame_period(hud : FFDHud):
	frame_period = 0.0
	def set_frame_period(period, menu):
		nonlocal frame_period
		frame_period = period
		menu.hide()
	menu = FFDHudMenu(hud, "Choose Difficulty")
	menu.add_option("Easy", partial(set_frame_period, 0.5, menu))
	menu.add_option("Medium", partial(set_frame_period, 0.25, menu))
	menu.add_option("Hard", partial(set_frame_period, 0.125, menu))
	menu.add_option("Insane", partial(set_frame_period, 0.0625, menu))
	menu.show()
	return frame_period

def main(hud_dev : str = "45D_HUD0"):
	HUD = FFDHud("/dev/" + hud_dev)
	game = Game(HUD)
	game.loop(get_frame_period(HUD))

if __name__ == "__main__":
	main()
